unit Frm5250;

{
 Telnet 5250 emulation (C) 1999-2005, by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TField = class;

  TFrame5250 = class(TFrame)
  private
    { Dclarations prives }
   fCanvas     :TCanvas;
   fCols       :integer;    // 80
   fRows       :integer;    // 25
   fFontHeight :integer;    // Canvas.TextHeight('A')
   fFontWidth  :integer;    // canvas.TextWith('W')
   fStatusRect :TRect;
   fStatusColor:TColor;     // Status Bar Color
   fStatus     :string;     // status text
   fLock       :boolean;
   fTextAttr   :byte;       // 5250 attribute
   fTextColor  :TColor;     // Windows TextColor
   fBackColor  :TColor;     // Windows BackColor
   fFontStyle  :TFontStyles;// [fsUnderline] of []
   fInsert     :boolean;
   fWhereX     :integer;
   fWhereY     :integer;
   fHome       :boolean;
   fHomeX      :integer;
   fHomeY      :integer;
   fShowAttr   :boolean;
   fMemSize    :integer;
   fMem        :pchar;      // Array[0..fRows-1,0..fCols-1] of char;
   fAtt        :pchar;      // Array[0..fRows-1,0..fCols-1] of byte;
   fErrLine    :integer;
   fBlink      :TColor;
   fBlue       :TColor;
   fGreen      :TColor;
   fPink       :TColor;
   fRed        :TColor;
   fTurquoise  :TColor;
   fWhite      :TColor;
   fYellow     :TColor;
   fOpCode     :byte;
   fFlags      :word;
   fStream     :string;
   fIndex      :integer;
   fReply      :string;
   fUpdate     :integer;
   fFields     :TList;
   fField      :TField;
   fScreens    :TList; // of TScreenShoot
   fClipboard  :TStringList;
   EOnSend     :TNotifyEvent;
   procedure WMEraseBkGnd(var Msg:TWMEraseBkGnd); message wm_erasebkgnd;
   procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
   procedure Paint; virtual;
   procedure InvalidateChar(x,y:integer);
   procedure SetStatus(Value:string);
   procedure SetTextAttr(Value:byte);
   procedure SetSize(ACols,ARows:integer);
   function WhereX:integer;
   function WhereY:integer;
   function CharX(x:integer):integer;
   function CharY(y:integer):integer;
   procedure SetOpCode(Value:Byte);
   procedure SetStream(Value:string);
   function GetChar:char;
   function GetByte:byte;
   function GetWord:word;
   function GetString:string;
   function GetCardinal:cardinal;
   procedure BeginUpdate;
   procedure EndUpdate;
   procedure SendReply;
   procedure SaveScreen;
   procedure RestoreScreen;
   procedure WriteToDisplay;
   procedure WriteStructuredField;
   procedure StartOfHeader;
   procedure StartOfField;
   procedure RepeatToAddress;
   procedure ClearUnitAlternate;
   procedure ReadScreenImmediat;
   procedure WriteErrorCode;
   procedure Write(s:string);
   procedure SetChar(x,y:integer; c:char);  // update fMem, TextAttr and call DrawChar
   function  PutChar(x,y:integer; c:char):boolean;  // Try to update a Field (EBCDIC)
   function  PutKey(x,y:integer; c:char):boolean; // Try to update a Field (ANSI)
   procedure ClrScr;
   procedure ClearFields;
   procedure NextField;
   procedure PrevField;
   function FindFieldXY(x,y:integer):boolean;
   procedure GotoXY(x,y:integer);
   procedure MoveTo(x,y:integer);
   procedure SystemRequest;
   procedure ATTNRequest;
   procedure Query5250;
   procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
   procedure WMKillFocus(Var Message: TWMKillFocus); message WM_KILLFOCUS;
   procedure WMGetDlgCode(Var Msg:TWMGetDlgCode); message wm_getdlgcode;
   procedure WMKeyDown(Var Msg:TWMKeyDown); message wm_keydown;
   procedure BackSpace;
   procedure Submit;
   procedure SendFields(AID_CODE:byte);
   property TextAttr:byte read fTextAttr write SetTextAttr;
  protected
   procedure CreateParams(var Params: TCreateParams); override;
   procedure PaintWindow(DC: HDC); override;
   procedure Resize; override;
   procedure KeyPress(var Key:char); override;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    { Dclarations publiques }
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;
   procedure CopyAsHTML(Html:TStrings);
   function AsHtml:string;
   function AsText:string;
   procedure Lock;
   procedure UnLock;
   procedure CopyFields;
   procedure PastFields;
   procedure CancelInvite;
   property Status:string read fStatus write SetStatus;
   property Blink:TColor read fBlink write fBlink;
   property Blue :TColor read fBlue write fBlue;
   property Green:TColor read fGreen write fGreen;
   property Pink :TColor read fPink write fPink;
   property Red  :TColor read fRed write fRed;
   property Turquoise:TColor read fTurquoise write fTurquoise;
   property White:TColor read fWhite write fWhite;
   property Yellow:TColor read fYellow write fYellow;
   property ShowAttr:boolean read fShowAttr write fShowAttr;
   property OpCode:byte read fOpCode write SetOpCode;
   property Flags:word read fFlags write fFlags;
   property Stream:string read fStream write SetStream;
   property OnSend:TNotifyEvent read EOnSend write EOnSend;
   property Reply:string read fReply;
  end;

  TField = class
  private
   use:integer;
   Frame:TFrame5250;
   Top,Left:integer;
   FFW,FCW,FLL:word;
   fText:string;
   fPos:integer; // fix #2
   fUpdated:boolean;
   l1,l2:integer;
   Constructor Create(AOwner:TFrame5250; X,Y:integer; aFFW,aFCW,aFLL:word);
   procedure release;
   function Own(X,Y:integer):boolean;
   function EmptyPos:integer;
   function Mandatory:boolean;
   procedure PutKey(c:char);  // ANSI
   procedure PutChar(c:char); // EBCDIC
   procedure Letchar(x:integer; c:char); // update DSP also
   Procedure EndOfField;
   procedure Home;
   procedure GotoEnd;
   procedure DelChar;
   procedure Past(s:string);
  end;


implementation

{$R *.dfm}

uses
 EBCDIC;

const
 eSPACE=#$40;

procedure digits_only(var c:char);
begin
 if not (c in ['0'..'9']) then c:=#0;
end;

procedure signed_numeric(var c:char);
begin
 if not (c in ['+','-']) then digits_only(c);
end;

procedure numeric_only(var c:char);
begin
 if pos(c,'+-,. ')=0 then digits_only(c);
end;

procedure alpha_only(var c:char);
begin
 if not (upcase(c) in ['A'..'Z']) then numeric_only(c);
end;

procedure Error(s:string);
begin
 raise Exception.Create(s);
end;

procedure SetFontSize(Font:TFont; Width,Height:integer);
var
 lf:TLogFont;
begin
 GetObject(Font.handle,sizeof(lf),@lf);
 lf.lfWidth :=Width;
 lf.lfHeight:=Height;
 Font.handle:=CreateFontIndirect(lf);
end;

function ccsColor(c:TColor):string;
var
 i:integer;
begin
 i:=(c and $0000FF) shl 16+ (c and $00FF00) + (c shr 16);
 result:='#'+inttohex(i,6);
end;

function ccsFontStyle(name:string; c1,c2:TColor):string;
var
 s1,s2:string;
begin
 s1:=ccsColor(c1);
 s2:=ccsColor(c2);
 result:='.'+name+'  { color:'+s1+'; background-color:'+s2+'}'#13#10+
         '.r'+name+' { color:'+s2+'; background-color:'+s1+'}';
end;

{$R SYSREQ.RES}

function GetDlgText(HWnd,Item:integer):string; // ok
var
 Text:array[0..50] of char;
begin
 GetDlgItemText(HWnd,Item,Text,50);
 Result:=Text;
end;

function SysReq(HWnd,Msg,wParam,lParam:integer):integer; stdcall; // ok
var
 Frame:TFrame5250;
begin
 Result:=1;
 case Msg of
  WM_INITDIALOG:begin
   SetProp(HWnd,'DSP5250',lParam);
   Frame:=pointer(lParam);
   SetDlgItemText(HWnd,101,PChar(Frame.fReply));
  end;
  WM_COMMAND:
   case wParam of
     1: begin  // OK
         Frame:=Pointer(GetProp(HWnd,'DSP5250'));
         Frame.fReply:=GetDlgText(HWnd,101);
         RemoveProp(HWnd,'DSP5250');
         EndDialog(HWnd,1);
        end;
     2: begin
         RemoveProp(HWnd,'DSP5250');
         EndDialog(HWnd,2);
        end;
   else Result:=DefWindowProc(Hwnd,Msg,wParam,lParam);
   end;
 else
  Result:=DefWindowProc(Hwnd,Msg,wParam,lParam);
 end;
end;

{ TField }
Constructor TField.Create(AOwner:TFrame5250; X,Y:integer; aFFW,aFCW,aFLL:word);
begin
 use:=1;
 Frame:=AOwner;
 Left:=X;
 Top:=Y;
 FFW:=aFFW;
 FCW:=aFCW;
 FLL:=aFLL;
 l1:=Left+Frame.fCols*Top; // Graham fix
 l2:=l1+FLL;
 Frame.fFields.Add(Self);
 SetLength(fText,FLL);
 FillChar(fText[1],FLL,eSPACE);
 fUpdated:=False;
 Frame.Write(fText);
end;

procedure TField.Release;
begin
 dec(use);
 if use<=0 then free;
end;

function TField.Own(X,Y:integer):boolean;
var
 ll:integer;
begin
 ll:=(x+Frame.fCols*y); // Graham fix
 Result:=(ll>=l1)and(ll<l2);
 if Result then fPos:=ll-l1+1;
end;

function TField.EmptyPos:integer;
var
 i:integer;
begin
 Result:=0;
 for i:=1 to FLL do if fText[i]=eSPACE then inc(Result);
end;

function TField.Mandatory:boolean;
begin
 case (FFW and $000F) of
  7: Result:=(EmptyPos>0);   // Mandatory Fill  (MF)
  8: Result:=(EmptyPos=FLL); // Mandatory Enter (ME)
  else Result:=False;
 end;
 if Result then begin
  Frame.MoveTo(Left+1,Top+1);
  Frame.Status:='Mandatory field';
  Beep;
 end;
end;

procedure TField.PutKey(c:char);
var
 Next:boolean;
 i:integer;
begin
// bit 10 (reverse order) = monocase
 if (FFW and $0020)>0 then c:=Upcase(c);
// bits 5-7 (reverse order) = Alpha, Number...
 case (FFW and $0700) of //
  $0100 : alpha_only(c);
  $0200 : { numeric shift };
  $0300 : numeric_only(c);
  $0400 : { katakana };
  $0500 : digits_only(c);   { digits only }
  $0600 : c:=#0; // io feature
  $0700 : signed_numeric(c);{ signed numeric }
 end;
 Next:=(fPos=Length(fText));
// Graham hide the caret...
 if Frame.fInsert then for i:=FLL downto fPos+1 do LetChar(i,fText[i-1]);
 PutChar(A2E[c]);
 fUpdated:=True;
// check also ErrLock
 if Next then begin
  if (FFW and $0080)>0 then // auto enter
   Frame.Submit
  else
   Frame.NextField;
 end;
end;

procedure TField.PutChar(c:char);
begin
 if c=#0 then c:=eSPACE;
 fText[fPos]:=c;
// many more things by Graham :)
 Frame.Write(c);
end;

procedure TField.LetChar(x:integer; c:char);
var
 y:integer;
begin
 fText[x]:=c;
 inc(x,Left-1);
 y:=Top;
 while x>=Frame.fCols do begin
  dec(x,Frame.fCols);
  inc(y);
 end;
 Frame.SetChar(x,y,c);
end;

procedure TField.Past(s:string);
var
 i,j:integer;
begin
 fUpdated:=True;
 j:=Length(s);
 if j>FLL then j:=FLL;
 for i:=1 to j do LetChar(i,s[i]);
end;

procedure TField.EndOfField;
var
 x:integer;
 ll:integer;
 i:integer;
begin
 Frame.BeginUpdate;
 fUpdated:=True; // fix 20/01/2000
 ll:=Frame.fWhereX+(Frame.fCols*Frame.fWhereY)+1;
 i:=ll-l1;
 for x:=ll to l2 do begin
  fText[i]:=eSPACE;
  inc(i);
  Frame.Write(eSPACE);
 end;
 Frame.NextField;
 Frame.EndUpdate;
end;

procedure TField.Home;
begin
 Frame.GotoXY(l1 mod Frame.fCols+1,l1 div Frame.fCols+1);
end;

procedure TField.GotoEnd;
begin
 Frame.fWhereX:=l2 mod Frame.fCols;
 Frame.fWhereY:=l2 div Frame.fCols;
end;

procedure TField.DelChar;
var
 i:integer;
begin
 fUpdated:=True;
 for i:=fPos to FLL-1 do LetChar(i,fText[i+1]);
 LetChar(FLL,eSPACE);
end;

{ TScreenShoot }
type
 TScreenShoot=class
 private
  fFrame :TFrame5250;
  fCols  :integer;
  fRows  :integer;
  fSize  :integer;
  fAtt   :PChar;
  fMem   :PChar;
  fFields:TList;
  fWhereX:integer;
  fWhereY:integer;
 public
  constructor Create(Frame:TFrame5250);
  destructor destroy; override;
  procedure Restore;
 end;

constructor TScreenShoot.Create(Frame:TFrame5250);
var
 i:integer;
begin
 fFrame:=Frame;
 fFrame.fScreens.Add(Self);
 fSize:=fFrame.fMemSize;
 fCols:=fFrame.fCols;
 fRows:=fFrame.fRows;
 GetMem(fMem,fSize);
 GetMem(fAtt,fSize);
 Move(fFrame.fMem^,fMem^,fSize);
 Move(fFrame.fAtt^,fAtt^,fSize);
 fFields:=TList.Create;
 for i:=0 to fFrame.fFields.Count-1 do begin
  fFields.Add(fFrame.fFields[i]);
  inc(TField(fFields[i]).Use);
 end;
 fWhereX:=fFrame.fWhereX;
 fWhereY:=fFrame.fWhereY;
end;

destructor TScreenShoot.destroy;
var
 i:integer;
begin
 FreeMem(fMem);
 FreeMem(fAtt);
 for i:=0 to fFields.Count-1 do TField(fFields[i]).Release;
 fFields.Free;
 fFrame.fScreens.Remove(Self);
 inherited;
end;

procedure TScreenShoot.Restore;
var
 i:integer;
begin
 fFrame.fCols:=fCols;
 fFrame.fRows:=fRows;
 if fFrame.fMemSize<>fSize then begin
  FreeMem(fFrame.fMem);
  FreeMem(fFrame.fAtt);
  fFrame.fMemSize:=fSize;
  GetMem(fFrame.fMem,fSize);
  GetMem(fFrame.fAtt,fSize);
 end;
 Move(fMem^,fFrame.fMem^,fSize);
 Move(fAtt^,fFrame.fAtt^,fSize);
 fFrame.ClearFields;
 for i:=0 to fFields.Count-1 do begin
  fFrame.fFields.Add(fFields[i]);
 end;
 fFrame.fWhereX:=fWhereX;
 fFrame.fWhereY:=fWhereY;
 fFrame.Resize;
end;

{ TFrame5250 }

constructor TFrame5250.Create(AOwner:TComponent);
begin
 inherited;
 FCanvas := TControlCanvas.Create;
 TControlCanvas(FCanvas).Control := Self;
 fCols:=80;
 fRows:=25;
 fStatusColor:=clBlack;
 fMemSize:=fCols*fRows;
 fMem:=AllocMem(fMemSize);
 fAtt:=AllocMem(fMemSize);
 fBlink :=clMaroon;
 fBlue  :=clBlue;
 fGreen :=clGreen;
 fPink  :=clFuchsia;
 fRed   :=clRed;
 fTurquoise:=clAqua;
 fWhite :=clWhite;
 fYellow:=clYellow;
 fTextAttr:=$20;
 fStatusColor:=clBlack;
// fStatusText:=clWhite;
 fStatus:='Offline';
 fScreens:=TList.Create;
 fFields:=TList.Create;
end;

destructor TFrame5250.Destroy;
begin
 ClearFields;
 fFields.Clear;
 while fScreens.Count>0 do TScreenShoot(fScreens.Last).Free;
 fScreens.Free;
  FCanvas.Free;
 inherited;
end;

procedure TFrame5250.CopyAsHTML(Html:TStrings);
const   //123456789
 escapes='"&<>';
 codes:array[1..9] of pchar=(
  'quot','amp','lt','gt',
  'agrave',
  'egrave','eacute','ecirc','euml'
 );
var
 body:string;
 x,y,z:integer;
 la:char;
 a:char;
 r:boolean;
 u:boolean;
 na:char;
 nr:boolean;
 nu:boolean;
 c:char;
 b:byte;
 fu:boolean;
 i:integer;
begin
 Html.Add('<!DOCTYPE HTML SYSTEM>');
 Html.Add('<HTML><HEAD><TITLE>AS/400</TITLE><STYLE>');

 html.Add(ccsFontStyle('g',fGreen,Color));
 html.Add(ccsFontStyle('w',fWhite,Color));
 html.Add(ccsFontStyle('r',fRed  ,Color));
 html.Add(ccsFontStyle('k',fBlink,Color));
 html.Add(ccsFontStyle('t',fTurquoise,Color));
 html.Add(ccsFontStyle('y',fYellow,Color));
 html.Add(ccsFontStyle('p',fPink,Color));
 html.Add(ccsFontStyle('b',fBlue,Color));
 html.Add('</STYLE></HEAD><BODY>');

 body:='<table><pre class="g">';

 la:=#$20; // last attribut
 a:='-';  // current attribut
 r:=false; // reverse
 u:=false; // underline
 na:='g';   // next attribut
 nr:=false; // next reverse
 nu:=false; // next underline

 z:=0;

 for y:=0 to fRows-1 do begin
  fu:=false;
  for x:=0 to fCols-1 do begin
   c:=fMem[z];
   case c of
    #$20..#$3F: begin // attribut

     if c<>la then begin // new attribut
      la:=c;
      b:=ord(c);
      if (b and 7)=7 then b:=0;
      nu:=((b and 4)<>0);
      case b and $3A of
       $20: na:='g';
       $22: na:='w';
       $28: na:='r';
       $2A: na:='k';
       $30: na:='t';
       $32: na:='y';
       $38: na:='p';
       $3A: na:='b';
       else na:='-';
      end;
      nr:=odd(b);
     end;

     if r and (a<>'-') then begin
      body:=body+'</font>';
      a:='-';
      r:=false;
     end;
     if u then begin
      body:=body+'</u>';
      u:=false;
     end;
     body:=body+' ';

    end;
    else begin
     if c=#0 then nu:=false;
     if (na='-')or(c=#0) then c:=' ' else c:=E2A[c];
     if (u<>nu) or (r<>nr) or ((a<>na)and(c<>' ')) then begin
      if (a='-')or((a='g')and(r=false)) then else body:=body+'</font>';
      a:=na;
      r:=nr;
      if u<>nu then if nu then body:=body+'<u>' else body:=body+'</u>';
      u:=nu;
      if (a='-')or((a='g')and(r=false)) then
      else begin
       body:=body+'<font class="';
       if r then body:=body+'r';
       body:=body+a+'">';
      end;
     end;
     if u then fu:=true;
     i:=pos(c,escapes);
     if i>0 then
      body:=body+'&'+codes[i]+';'
     else begin
      if (c=' ')and(fu and (x=fCols-1)) then body:=body+'&nbsp;' // force underline on space
      else body:=body+c;
     end;
     (*
     case c of
      '<' : body:=body+'&lt;';
      '>' : body:=body+'&gt;';
      '&' : body:=body+'&amp;';
      '' : body:=body+'&eacute;';
      '' : body:=body+'&egrave;';
     end;
     *)
    end;
   end;

   inc(z);
  end;
  html.add(body);
  body:='';
 end;

 html.Add('</font></pre></table>');

 Html.Add('TN400, &copy; 1999-2005 by Paul TOTH &lt;<a href="mailto:tothpaul@free.fr">tothpaul@free.fr</a>&gt;<br>');
 Html.Add('<a href="http://tothpaul.free.fr">http://tothpaul.free.fr</a>');

 Html.Add('</BODY></HTML>');
end;

function TFrame5250.AsHTML:string;
const   //123456789
 escapes='"&<>';
 codes:array[1..9] of pchar=(
  'quot','amp','lt','gt',
  'agrave',
  'egrave','eacute','ecirc','euml'
 );
var
 html:TStringList;
 body:string;
 x,y,z:integer;
 la:char;
 a:integer;
 r:boolean;
 u:boolean;
 na:integer;
 nr:boolean;
 nu:boolean;
 c:char;
 b:byte;
 fu:boolean;
 i:integer;
 styles:array[False..True,0..7] of string;
begin
 Html:=TStringList.Create;
 try
  Html.Add('<HTML><HEAD><TITLE>AS/400</TITLE></HEAD>');
  styles[False,0]:='color:'+ccsColor(fGreen)+'; background-color:'+ccsColor(Color);
  styles[False,1]:='color:'+ccsColor(fWhite)+'; background-color:'+ccsColor(Color);
  styles[False,2]:='color:'+ccsColor(fRed  )+'; background-color:'+ccsColor(Color);
  styles[False,3]:='color:'+ccsColor(fBlink)+'; background-color:'+ccsColor(Color);
  styles[False,4]:='color:'+ccsColor(fTurquoise)+'; background-color:'+ccsColor(Color);
  styles[False,5]:='color:'+ccsColor(fYellow)+'; background-color:'+ccsColor(Color);
  styles[False,6]:='color:'+ccsColor(fPink)+'; background-color:'+ccsColor(Color);
  styles[False,7]:='color:'+ccsColor(fBlue)+'; background-color:'+ccsColor(Color);
  styles[True,0]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fGreen);
  styles[True,1]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fWhite);
  styles[True,2]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fRed);
  styles[True,3]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fBlink);
  styles[True,4]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fTurquoise);
  styles[True,5]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fYellow);
  styles[True,6]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fPink);
  styles[True,7]:='color:'+ccsColor(Color)+'; background-color:'+ccsColor(fBlue);
 (* Copy/Past doesn't support header styles :(
  html.Add(ccsFontStyle('g',fGreen,Color));
  html.Add(ccsFontStyle('w',fWhite,Color));
  html.Add(ccsFontStyle('r',fRed  ,Color));
  html.Add(ccsFontStyle('k',fBlink,Color));
  html.Add(ccsFontStyle('t',fTurquoise,Color));
  html.Add(ccsFontStyle('y',fYellow,Color));
  html.Add(ccsFontStyle('p',fPink,Color));
  html.Add(ccsFontStyle('b',fBlue,Color));
 *)
  html.Add('<BODY><!--StartFragment-->');

  body:='<table><font style="font-family: Courrier; font-size: 10px"><pre style="'+styles[false,0]+'">';

  la:=#$20; // last attribut
  a:=8;  // current attribut
  r:=false; // reverse
  u:=false; // underline
  na:=0;   // next attribut
  nr:=false; // next reverse
  nu:=false; // next underline

  z:=0;

  for y:=0 to fRows-1 do begin
   fu:=false;
   for x:=0 to fCols-1 do begin
    c:=fMem[z];
    case c of
     #$20..#$3F: begin // attribut

      if c<>la then begin // new attribut
       la:=c;
       b:=ord(c);
       if (b and 7)=7 then b:=0;
       nu:=((b and 4)<>0);
       case b and $3A of
        $20: na:=0;
        $22: na:=1;
        $28: na:=2;
        $2A: na:=3;
        $30: na:=4;
        $32: na:=5;
        $38: na:=6;
        $3A: na:=7;
        else na:=8;
       end;
       nr:=odd(b);
      end;

      if r and (a<>8) then begin
       body:=body+'</font>';
       a:=8;
       r:=false;
      end;
      if u then begin
       body:=body+'</u>';
       u:=false;
      end;
      body:=body+' ';

     end;
     else begin
      if c=#0 then nu:=false;
      if (na=8)or(c=#0) then c:=' ' else c:=E2A[c];
      if (u<>nu) or (r<>nr) or ((a<>na)and(c<>' ')) then begin
       if (a=8)or((a=0)and(r=false)) then else body:=body+'</font>';
       a:=na;
       r:=nr;
       if u<>nu then if nu then body:=body+'<u>' else body:=body+'</u>';
       u:=nu;
       if (a=8)or((a=0)and(r=false)) then
       else begin
        body:=body+'<font style="'+styles[r,a]+'">';
       end;
      end;
      if u then fu:=true;
      i:=pos(c,escapes);
      if i>0 then
       body:=body+'&'+codes[i]+';'
      else begin
       if (c=' ')and(fu and (x=fCols-1)) then body:=body+'&nbsp;' // force underline on space
       else body:=body+c;
      end;
     end;
    end;

    inc(z);
   end;
   html.add(body);
   body:='';
  end;

  html.Add('</font></pre></font></table>');

  Html.Add('TN400, &copy; 1999-2005 by Paul TOTH &lt;<a href="mailto:tothpaul@free.fr">tothpaul@free.fr</a>&gt;<br>');
  Html.Add('<a href="http://tothpaul.free.fr">http://tothpaul.free.fr</a>');

  Html.Add('<!--EndFragment--></BODY></HTML>');
  Result:=Html.Text;
 finally
  Html.Free;
 end;
end;

function TFrame5250.AsText:string;
var
 x,y,z:integer;
 c:char;
 i:integer;
begin
 SetLength(Result,fRows*(fCols+2));
 z:=0;
 i:=1;
 for y:=0 to fRows-1 do begin
  for x:=0 to fCols-1 do begin
   c:=fMem[z];
   case c of
    #$00,
    #$20..#$3F: c:=' '; // attribut
    else        c:=E2A[c];
   end;
   Result[i]:=c; inc(i);
   inc(z);
  end;
  Result[i]:=#13; inc(i);
  Result[i]:=#10; inc(i);
 end;
// Result:=Result+'TN400, (c) 1999-2005 by Paul TOTH <tothpaul@free.fr>';
end;

procedure TFrame5250.WMEraseBkGnd(var Msg: TWMEraseBkGnd);
begin
// vite le clignotement
 Msg.Result:=1;
end;

procedure TFrame5250.WMPaint(var Message: TWMPaint);
begin
 ControlState:=ControlState+[csCustomPaint];
 inherited;
 ControlState:=ControlState-[csCustomPaint];
end;

procedure TFrame5250.PaintWindow(DC: HDC);
begin
  FCanvas.Lock;
  try
    FCanvas.Handle := DC;
    try
      TControlCanvas(FCanvas).UpdateTextFlags;
      Paint;
    finally
      FCanvas.Handle := 0;
    end;
  finally
    FCanvas.Unlock;
  end;
end;

procedure TFrame5250.Paint;
var
 r:TRect;
 a,c:char;
 l1,l2,c1,c2:integer;
 x,y,o:integer;
 s:string;
 last:char;
begin
 r:=fCanvas.ClipRect;
// status line
 if r.Bottom>=fStatusRect.Top then begin
  fCanvas.Brush.Color:=fStatusColor;
  fCanvas.FillRect(fStatusRect);
  x:=fFontWidth div 2;
  y:=fStatusRect.Top;
  TextAttr:=$22; // White
  fCanvas.TextOut(x,y,fStatus);
  if fInsert then fCanvas.TextOut(Width div 2,y,'I');
  s:=IntToStr(WhereX)+'/'+IntToStr(WhereY);
  fCanvas.TextOut(Width-fCanvas.TextWidth(s)-x,y,s);
  dec(y);
  fCanvas.Pen.Color:=clWhite;
  fCanvas.MoveTo(0,y);
  fCanvas.LineTo(Width,y);
 end;

 l1:=0; while CharY(l1+1)<r.Top do inc(l1);
 l2:=l1-1; while CharY(l2+1)<r.Bottom do inc(l2);
 c1:=0; while CharX(c1+1)<r.Left do inc(c1);
 c2:=c1-1; while CharX(c2+1)<r.Right do inc(c2);
 if l1<0 then l1:=0 else if l1>=fRows then l1:=fRows-1;
 if l2<0 then l2:=0 else if l2>=fRows then l2:=fRows-1;
 if c1<0 then c1:=0 else if c1>=fCols then c1:=fCols-1;
 if c2<0 then c2:=0 else if c2>=fCols then c2:=fCols-1;

 a:=fAtt[c1+fCols*l1];
 TextAttr:=ord(a);
 last:=#0;
 for y:=l1 to l2 do begin
  for x:=c1 to c2 do begin
   o:=x+y*fCols;
   if fAtt[o]<>a then begin
    a:=fAtt[o];
    TextAttr:=ord(a);
   end;
   r.Top :=CharY(y);
   r.Left:=CharX(x);
   r.Right :=r.Left+fFontWidth +1;
   r.Bottom:=r.Top +fFontHeight+1;
   c:=fMem[o];
   case c of
    #$00,
    #$20..#$3F: begin
     if c<>#0 then fCanvas.Brush.Color:=Color;
     fCanvas.FillRect(r);
     if fShowAttr and ((c<>#0)or(c<>last)) then begin
      last:=c;
      TextAttr:=$22;
      SetFontSize(fCanvas.Font,fFontWidth div 2,fFontHeight div 2);
      fCanvas.TextOut(r.Left,r.Top,IntToHex(ord(c) div 16,1));
      fCanvas.TextOut(r.Left+fFontWidth div 2,r.Top+fFontHeight div 2,IntToHex(ord(c) and 15,1));
      TextAttr:=ord(a);
     end;
    end;
    else begin
     fCanvas.Brush.Color:=fBackColor;
     fCanvas.TextRect(r,r.Left,r.Top,E2A[c]);
    end;
   end;
  end;
 end;
end;

procedure TFrame5250.CreateParams(var Params: TCreateParams);
begin
 inherited;
 with Params.WindowClass do style := style + CS_VREDRAW + CS_HREDRAW;
end;

procedure TFrame5250.Resize;
begin
 inherited;
 fFontWidth :=ClientWidth  div (fCols);
 fFontHeight:=(ClientHeight-2) div (fRows+1);
 fStatusRect:=Rect(0,CharY(fRows)+2,Width,Height);
 if Focused then begin
  DestroyCaret;
  CreateCaret(Handle,0,fFontWidth,fFontHeight);
  SetCaretPos(CharX(fWhereX),CharY(fWhereY));
  if not fLock then ShowCaret(Handle);
 end;
end;

procedure TFrame5250.SetTextAttr(Value: byte);
begin
 if (Value and 7)=7 then Value:=0; // not displayed attributs
 fTextAttr:=Value;
 if (fTextAttr and $4)=0 then fFontStyle:=[] else fFontStyle:=[fsUnderline]; // Underline
 case fTextAttr and $3A of // color
  $20: fTextColor:=fGreen;
  $22: fTextColor:=fWhite;
  $28: fTextColor:=fRed;
  $2A: fTextColor:=fBlink;
  $30: fTextColor:=fTurquoise;
  $32: fTextColor:=fYellow;
  $38: fTextColor:=fPink;
  $3A: fTextColor:=fBlue;
  else fTextcolor:=Color; // invisible !
 end;
 if odd(fTextAttr) then begin // Reverse
  fBackColor:=fTextColor;
  fTextColor:=Color;
 end else begin
  fBackColor:=Color;
 end;
 with fCanvas do begin
  Font:=Self.Font;
  Font.Style:=fFontStyle;
  Font.Color:=fTextColor;
  Brush.Color:=fBackColor;
  SetFontSize(Font,fFontWidth,fFontHeight);
 end;
end;

function TFrame5250.WhereX: integer;
begin
 Result:=fWhereX+1;
end;

function TFrame5250.WhereY: integer;
begin
 Result:=fWhereY+1;
end;

function TFrame5250.CharX(x: integer): integer;
begin
 Result:=(x*Width) div fCols;
end;

function TFrame5250.CharY(y: integer): integer;
begin
 Result:=(y*Height) div (fRows+1);
end;

procedure TFrame5250.SetStatus(Value: string);
begin
 if fStatus<>Value then begin
  fStatus:=Value;
  InvalidateRect(Handle,@fStatusRect,True);
 end;
end;

procedure TFrame5250.Lock;
begin
 if fLock then exit;
 fLock:=True;
 HideCaret(Handle);
 Status:='Busy';
end;

procedure TFrame5250.SetOpCode(Value: Byte);
begin
 fOpCode:=Value;
 case fOpCode of
  01 : UnLock;
  10 : CancelInvite;
  12 : {Message Off};
 end;
end;

procedure TFrame5250.UnLock;
begin
 Status:='OK';
 if Focused then ShowCaret(Handle) else Beep;
 fLock:=False;
end;

procedure TFrame5250.CopyFields;
var
 i:integer;
begin
 if fClipBoard=nil then fClipBoard:=TStringList.Create else fClipBoard.Clear;
 for i:=0 to fFields.count-1 do fClipBoard.Add(TField(fFields[i]).fText);
end;

procedure TFrame5250.PastFields;
var
 i,j:integer;
begin
 if fClipBoard=nil then exit;
 j:=fClipBoard.Count-1;
 if j>=fFields.Count then j:=fFields.Count-1;
 for i:=0 to j do begin
  TField(fFields[i]).Past(fClipboard[i]);
 end;
end;

procedure TFrame5250.CancelInvite;
begin
 Lock; // ?
 fReply:='';
 fOpCode:=10;
 fFlags:=0;
 SendReply;
end;

procedure TFrame5250.SendReply;
begin
 if Assigned(EOnSend) then begin
  EOnSend(Self);
  Lock;
 end;
end;

procedure TFrame5250.SetStream(Value: string);
var
 c:char;
begin
 fStream:=Value;
 fIndex:=1;
 BeginUpdate;
 While fIndex<=Length(fStream) do begin
  c:=GetChar;
  if c<>#04 then Error('Stream error (04) : '+IntToHex(ord(c),4));
  c:=Getchar;
  case c of
   #$02 : SaveScreen;
   #$11 : WriteToDisplay;
   #$12 : RestoreScreen;
   #$20 : ClearUnitAlternate;
   #$21 : WriteErrorCode;
   #$40 : {Clear} SetSize(80,25);
   #$42 : { read input fields } inc(fIndex,2);
   #$50 : ClearFields;
   #$52 : { read mandatory fields } inc(fIndex,2);
   #$62 : ReadScreenImmediat;
   #$F3 : WriteStructuredField;
   else Error('Unknow code $'+IntToHex(ord(c),2));
  end;
 end;
 EndUpdate;
end;

procedure TFrame5250.BeginUpdate;
begin
 if fUpdate=0 then HideCaret(Handle); // se if necessary
 inc(fUpdate);
end;

procedure TFrame5250.EndUpdate;
begin
 if fUpdate>0 then dec(fUpdate);
 if fUpdate=0 then begin
  Resize;
  Invalidate;
 end;
end;

function TFrame5250.GetChar: char;
begin
 if fIndex>Length(fStream) then Error('Read after end of stream !');
 Result:=fStream[fIndex];
 inc(fIndex);
end;

function TFrame5250.GetByte: byte;
begin
 if fIndex>Length(fStream) then Error('Read after end of stream !');
 Result:=ord(fStream[fIndex]);
 inc(fIndex);
end;

function TFrame5250.GetWord:word;
begin
 Result:=GetByte;
 Result:=256*Result+GetByte;
end;

function TFrame5250.GetString:string;
var
 b:byte;
begin
 b:=GetByte;
{$ifdef log2}writeln('GetString : ',b,' chars');{$endif}
 Result:='';
 while b>0 do begin
  Result:=Result+GetChar;
  dec(b);
 end;
end;

function TFrame5250.GetCardinal:cardinal;
begin
 if fIndex+3>Length(fStream) then Error('Read after end of stream !');
 move(fStream[fIndex],Result,4);
 inc(fIndex,4);
end;

procedure TFrame5250.SaveScreen;
var
 SS:TScreenShoot;
begin
 SS:=TScreenShoot.Create(Self);
 fReply:=#$04#$12#0#0'????';
 Move(SS,fReply[5],4);
 fOpCode:=4;
 fFlags:=0;
 SendReply;
end;

procedure TFrame5250.Write(s: string);
var
 x : integer;
begin
 if s='' then exit;
 if fWhereX+Length(s)>fCols then begin
  x:=fCols-fWhereX;
  Write(Copy(s,1,x));
  Delete(s,1,x);
  Write(s);
 end else begin
  if fUpdate=0 then HideCaret(Handle);  // Graham added the fUpdate test
  for x:=1 to length(s) do begin
   SetChar(fWhereX,fWhereY,s[x]);
   inc(fWhereX);
  end;
  MoveTo(fWhereX+1,fWhereY+1);
  if fUpdate=0 then ShowCaret(Handle); // Graham added the fUpdate test
 end;
end;

procedure TFrame5250.NextField;
var
 x,y:integer;
begin
 if fFields.Count=0 then exit;
// Skip this Field
 if FindFieldXY(fWhereX,fWhereY) then fField.GotoEnd;
// same line
 for x:=fWhereX to fCols-1 do begin
  if FindFieldXY(x,fWhereY) then begin
   MoveTo(x+1,fWhereY+1);
   exit;
  end;
 end;
// Other lines
 y:=fWhereY;
 Repeat
  inc(fWhereY);
  if fWhereY=fRows then fWhereY:=0;
  for x:=0 to fCols-1 do begin
   if FindFieldXY(x,fWhereY) then begin
    MoveTo(x+1,fWhereY+1);
    exit;
   end;
  end;
 until fWhereY=y; // no match ! when All Fields are Output Only !!! fix 06-19-2K
// So, Where should I put the Cursor ?!
end;

procedure TFrame5250.PrevField;
var
 x,y:integer;
begin
 if fFields.Count=0 then exit;
// Skip this Field
  if FindFieldXY(fWhereX,fWhereY) then fWhereX:=fField.Left;
// same line
 for x:=1 to fWhereX do begin
  if FindFieldXY(fWhereX-x,fWhereY) then begin
   MoveTo(fField.Left+1,fWhereY+1);
   exit;
  end;
 end;
// Other lines
 y:=fWhereY;
 Repeat
  dec(fWhereY);
  if fWhereY<0 then fWhereY:=fRows-1;
  for x:=1 to fCols do begin
   if FindFieldXY(fCols-x,fWhereY) then begin
    MoveTo(fField.Left+1,fWhereY+1);
    exit;
   end;
  end;
 until fWhereY=y;
end;

procedure TFrame5250.SetChar(x, y: integer; c: char);
var
 old:char;
 ax,ay,z:integer;
begin
 z:=x+fCols*y;
 if fMem[z]=c then exit;
 old:=fMem[z];
 fMem[z]:=c;
 if c in [#$20..#$3F] then begin
  fAtt[z]:=c;
  ax:=x+1;
  ay:=y;
  z:=ax+fCols*ay;
  while ay<fRows do begin
   while ax<fCols do begin
    if fMem[z] in [#$20..#$3F] then begin ay:=100; break end;
    fAtt[z]:=c; inc(z);
    InvalidateChar(ax,ay);
    Inc(ax);
   end;
   ax:=0;
   inc(ay);
  end;
 end else
 if old in [#$20..#$3F] then begin
  if (x+y=0) then c:=#$00 else c:=fAtt[z-1];
  ax:=x;
  ay:=y;
  while ay<fRows do begin
   while ax<fCols do begin
    z:=ax+fCols*ay;
    if fMem[z] in [#$20..#$3F] then begin ay:=100; break end;
    fAtt[z]:=c;
    InvalidateChar(ax,ay);
    Inc(ax);
   end;
   ax:=0;
   inc(ay);
  end;
 end;
 InvalidateChar(x,y);
end;

function TFrame5250.PutChar(x,y:integer; c:char):boolean; // ok
begin
 Result:=FindFieldXY(x,y);
 if Result then fField.PutChar(c); // ,False
end;

function TFrame5250.PutKey(x,y:integer;c:char):boolean; // +/-
begin
// if fLock then result:=false...
 Result:=FindFieldXY(x,y);
 if Result then fField.PutKey(c) { else ErrorLock(NoInput) };
end;

procedure TFrame5250.GotoXY(x, y: integer);
begin
 if (X>0)and(X<=fCols)
 and(Y>0)and(Y<=fRows) then begin
  fWhereX:=X-1;
  fWhereY:=Y-1;
  if fUpdate=0 then begin
   InvalidateRect(Handle,@fStatusRect,False);
   If Focused then begin
    SetCaretPos(CharX(fWhereX),CharY(fWhereY));
   end;
  end;
 end;
end;

procedure TFrame5250.ClearFields;
var
 i:integer;
begin
 for i:=0 to fFields.Count-1 do TField(fFields[i]).release;
 fFields.Clear;
 fWhereX:=0;
 fWhereY:=0;
end;

procedure TFrame5250.WriteToDisplay;
var
 x,y:integer;
 code:char;
 ccc1:byte; // command control caracter 1
begin
 x:=fWhereX;
 y:=fWhereY;
 GetChar; // ccc0
 ccc1:=ord(GetChar);
 while (fIndex<=Length(fStream)) and (fStream[fIndex]<>#4) do begin
  code:=GetChar;
  case code of
   #$01 : StartOfHeader; // Graham
   #$02 : RepeatToAddress;
   #$03 : ShowMessage('code $03 Erase');
   #$10 : ShowMessage('Transparent Data order');
   #$11 : begin // goto xy
           fWhereY:=GetByte-1;
           fWhereX:=GetByte-1;
          end;
   #$12 : ShowMessage('Write Extended attribute color');
   #$13 : begin// InsertCursor
           fHome:=True;
           fHomeY:=GetByte;
           fHomeX:=GetByte;
          end;
   #$14 : ShowMessage('Move Cursor ?');
   #$15 : ShowMessage('Structured Field');
   #$1D : StartOfField;
   else   if Not PutChar(fWhereX,fWhereY,code) then Write(code); // send chars to Fields / or screen
  end;
 end;
 if (ccc1 and $04)>0 then Beep;   // Sound Alarm
 if (ccc1 and $08)>0 then Unlock;
 if (ccc1 and $08)=8 then begin
  if fHome then begin // "ccc1 and 8" and "InserCursor" can occur on to different WTD...
   GotoXY(fHomeX,fHomeY);
   fHome:=False;
  end else begin
   GotoXY(1,1);
   NextField;
  end;
 end else begin
  fWhereX:=x;
  fWhereY:=y;
 end;
end;

procedure TFrame5250.RestoreScreen;
var
 SS:TScreenShoot;
 i:integer;
begin
 GetByte;
 GetByte;
 cardinal(SS):=GetCardinal;
 i:=fScreens.IndexOf(SS);
 if fScreens.Count=0 then
  ShowMessage('<!> Empty screen stack !')
 else begin
  if i<0 then
   ShowMessage('<!> Screen List corrupted !')
  else begin
   SS.Restore;
   fScreens.Delete(i);
  end;
 end;
 EndUpdate;
 //Invalidate;
end;

procedure TFrame5250.ClearUnitAlternate;
var
 cl:char;
begin
 cl:=GetChar;
 case cl of
  #$00: SetSize(132,28);
  #$80: ClrScr;
  else ShowMessage('<!> Unknow ClearUnitAlternate parameter #'+IntToStr(ord(cl)));
 end;
end;

procedure TFrame5250.WriteErrorCode;
var
 x,y:integer;
 code:char;
begin
 x:=fWhereX;
 y:=fWhereY;
 fWhereX:=0;
 fWhereY:=fErrLine; // ErrorLine pos
 while (fIndex<=length(fStream))and(fStream[fIndex]<>#4) do begin
  code:=GetChar;
  case code of
   #$13 : begin// InsertCursor
           fHome:=True;
           fHomeY:=GetByte-1;
           fHomeX:=GetByte-1;
          end;
   else begin
    if Not PutChar(fWhereX,fWhereY,code) then Write(code); // send chars to Fields / or screen
   end;
  end;
 end;
 fWhereX:=x;
 fWhereY:=y;
 UnLock; // this emulator doesn't need Restore Key :)
end;

procedure TFrame5250.SetSize(ACols, ARows: integer);
begin
 Lock;
 ClearFields;
 fWhereX:=0;
 fWhereY:=0;
 FreeMem(fMem);
 FreeMem(fAtt);
 fCols:=ACols;
 fRows:=ARows;
 fMemSize:=fCols*fRows;
 fMem:=AllocMem(fMemSize);
 FAtt:=AllocMem(fMemSize);
 ReSize;
end;

procedure TFrame5250.ReadScreenImmediat;
begin
 Lock;
 SetLength(fReply,fMemSize);
 Move(fMem[0],fReply[1],fMemSize);
 fOpCode:=0;
 fFlags:=0;
 SendReply;
end;

procedure TFrame5250.WriteStructuredField;
var
 size:integer;
 wsf:string;
begin
 size:=GetWord;
 wsf:='';
 while size>2 do begin
  wsf:=wsf+GetChar;
  dec(size);
 end;
 if wsf=#$D9#$70#$00 then Query5250 else ShowMessage('Unknow WSF');
end;

procedure TFrame5250.MoveTo(x, y: integer);
begin
 if X>fCols then begin inc(Y); X:=1; end;
 if X<=0 then begin dec(Y); X:=fCols; end;
 if Y>fRows then Y:=1;
 if Y<=0 then Y:=fRows;
 GotoXY(x,y);
end;

function TFrame5250.FindFieldXY(x, y: integer): boolean;
var
 i:integer;
begin
 for i:=0 to fFields.Count-1 do begin
  fField:=TField(fFields[i]);
  Result:=((fField.FFW and $2000)=0) and fField.Own(x,y);
  if Result then exit;
 end;
 Result:=false;
 fField:=nil;
end;

procedure TFrame5250.InvalidateChar(x, y: integer);
var
 r:TRect;
 i:integer;
begin
 if fUpdate>0 then exit;
 r.Left:=CharX(x);
 r.Top :=CharY(y);
 r.Right:=r.Left+fFontWidth+1;
 r.Bottom:=r.Top+fFontHeight+1;
{$ifdef async}
 InvalidateRect(Handle,@r,False);
{$else}
 if fUpdate=0 then HideCaret(Handle);
 i:=x+fCols*y;
 TextAttr:=ord(fAtt[i]);
 fCanvas.TextRect(r,r.left,r.top,E2A[fMem[i]]);
 if fUpdate=0 then ShowCaret(Handle); 
{$endif}
end;

procedure TFrame5250.StartOfHeader;
var
 s:string;
begin
 s:=GetString;
 if length(s)>=4 then fErrLine:=ord(s[4])-1; // fix #1
 ClearFields;
end;

procedure TFrame5250.StartOfField;
var
 ffw:word;
 fat:char;
 fcw:word;
 fll:word;
 att:char;
begin
 ffw:=GetByte;
 if (ffw and $40)=0 then begin
  fat:=chr(ffw); // attribut
  ffw:=0;
  fcw:=0;
 end else begin
  ffw:=256*ffw+GetByte;
  fcw:=GetByte;
  if (fcw and $80)=0 then begin
   fat:=chr(fcw); // attribut
   fcw:=0;
  end else begin
   fcw:=256*fcw+GetByte;
   fat:=GetChar;
  end;
 end;
 fll:=GetWord;
{$IFDEF LOG}
 WriteLn('StartOfField (',fll,'-FFW=',IntToHex(ffw,4),'-FCW=',IntToHex(fcw,4),' $',IntToHex(ord(fat),2),')');
{$ENDIF}
 att:=fAtt[fWhereX+fCols*fWhereY];
 Write(fat);
// Graham
 if FindFieldXY(fWhereX,fWhereY) then begin
  fField.fUpdated:=False;
  fField.FFW:=ffw;
  fField.FCW:=fcw;
  Write(fField.fText);
 end else begin
  fField:=TField.Create(Self,fWhereX,fWhereY,ffw,fcw,fll);
 end;
 Write(att); // fix a bug on reverse fields...but I really don't know if it's ok :(
 fField.Home;
end;
procedure TFrame5250.RepeatToAddress;
var
 sx,sy,dx,dy:integer;
 ch:char;
begin
 dy:=getbyte-1;
 dx:=getbyte-1;
 ch:=getchar;
 repeat
  sx:=fWhereX;
  sy:=fWhereY;
  if not PutChar(sx,sy,ch) then Write(ch);
 until (sx=dx)and(sy=dy);
end;

procedure TFrame5250.ClrScr;
begin
 Lock;
 ClearFields;
 FillChar(fMem^,fMemSize,$00);
 FillChar(fAtt^,fMemSize,$00);
 TextAttr:=0;
end;

procedure TFrame5250.Query5250;
begin
 fReply:=#$00#$00+     // X/Y =0
         #$88+
         #$00#$3A+     // Size=58
         #$D9#$70#$80+ // Query5250 reply
         #$06#$00+     // Controller hardware clas
         #$01#$01#$00+ // Revision
         #0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0+
         #$01+         // 5250 Display or 5250 emulation
        {
         #$F5#$F2#$F5#$F1+    // DEV 5251
         #$F0#$F1#$F1+        // MODEL 011
        }
         #$F5+#$F2+#$F9+#$F2+    // DEV 5292
         #$F0+#$F0+#$F2+         // 2
{          #$F3#$F4#$F7#$F7+    // DEV 3477
         #$C6+#$C3+eSPACE+    // FC}
         #02#00#00+    // KEYBOARD
         #$00#$61#$50#$00+     // Serial Number
         #$01#$00+     // Maximum input fields
         #0#0#0+
         #$01#$10+     // Customization
         #0#0#0#0#0#0#0#0#0#0;
 fOpCode:=3;
 fFlags:=0;
 SendReply;
end;

procedure TFrame5250.WMSetFocus(var Message: TWMSetFocus);
begin
 inherited;
 Resize;
end;

procedure TFrame5250.WMKillFocus(Var Message:TWMKillFocus);
begin
 DestroyCaret;
 inherited;
end;

procedure TFrame5250.WMGetDlgCode(Var Msg:TWMGetDlgCode);
begin
 inherited;
 Msg.Result:=Msg.Result or dlgc_WantAllKeys or dlgc_WantArrows or dlgc_WantTab;
end;

procedure TFrame5250.WMKeyDown(var Msg: TWMKeyDown);
var
 First:boolean;
 r:TRect;

 function Shift:boolean;
 begin
  result:=GetKeyState(VK_SHIFT) < 0;
 end;

 function Ctrl:boolean;
 begin
  result:=GetKeyState(VK_CONTROL) < 0;
 end;

begin
 if (fLock)and(Msg.CharCode<>vk_escape) then exit;
 First:=(Msg.KeyData and (1 shl 30))=0; // disable auto repeat for AID Keys
 case Msg.CharCode of
  vk_up     : if Shift then SendFields($F4) else MoveTo(WhereX,WhereY-1);
  vk_down   : if Shift then SendFields($F5) else MoveTo(WhereX,WhereY+1);
  vk_left   : MoveTo(WhereX-1,WhereY);
  vk_right  : MoveTo(WhereX+1,WhereY);
  vk_end    : if FindFieldXY(fWhereX,fWhereY) then fField.EndOfField;
  vk_f1     : if First then if Shift then SendFields($B1) else SendFields($31);
  vk_f2     : if First then if Shift then SendFields($B2) else SendFields($32);
  vk_f3     : if First then if Shift then SendFields($B3) else SendFields($33);
  vk_f4     : if First then if Shift then SendFields($B4) else SendFields($34);
  vk_f5     : if First then if Shift then SendFields($B5) else SendFields($35);
  vk_f6     : if First then if Shift then SendFields($B6) else SendFields($36);
  vk_f7     : if First then if Shift then SendFields($B7) else SendFields($37);
  vk_f8     : if First then if Shift then SendFields($B8) else SendFields($38);
  vk_f9     : if First then if Shift then SendFields($B9) else SendFields($39);
  vk_f10    : if First then if Shift then SendFields($BA) else SendFields($3A);
  vk_f11    : if First then if Shift then SendFields($BB) else SendFields($3B);
  vk_f12    : if First then if Shift then SendFields($BC) else SendFields($3C);
  vk_return : if First then Submit;
  vk_prior  : SendFields($F4);
  vk_next   : SendFields($F5);
  vk_escape : if First then begin if Shift then SystemRequest else ATTNRequest; end;
  vk_tab    : if First then if Shift then PrevField else NextField;
  vk_delete : if FindFieldXY(fWhereX,fWhereY) then fField.DelChar;
  vk_home   : if FindFieldXY(fWhereX,fWhereY) then fField.Home;
  vk_insert : {if Shift then CopyFields else if Ctrl then PastFields else} begin
               fInsert:=not fInsert;
               r:=Rect(0,Height-fFontHeight,Width,Height);
               InvalidateRect(Handle,@r,True);
              end;
  else inherited;
 end;
end;

procedure TFrame5250.KeyPress(var Key:char);
begin
 if fLock then exit;
 case Key of
  #8 : BackSpace;
  #9 : {nothing, see KeyDown};
  #13: {SendFields($F1); // AID_ENTER};
  #27: {SystemRequest}
  else begin
   PutKey(fWhereX,fWhereY,Key);
  end;
 end;
end;

procedure TFrame5250.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 c,l:integer;
begin
 inherited;
 if Focused then begin
  c:=1; while CharX(c)<X do inc(c);
  l:=1; while CharY(l)<Y do inc(l);
  GotoXY(c,l)
 end else SetFocus;
end;

procedure TFrame5250.BackSpace;
begin
 MoveTo(fWhereX,WhereY);
 if FindFieldXY(fWhereX,fWhereY) then begin
  fField.PutChar(eSPACE);
  MoveTo(fWhereX,WhereY);
 end;
end;

procedure TFrame5250.Submit;
var
 i:integer;
begin
 for i:=0 to fFields.Count-1 do begin
  if TField(fFields[i]).Mandatory then exit;
 end;
 SendFields($F1);
end;

procedure TFrame5250.SendFields(AID_CODE: byte);
var
 i:integer;
begin
 SetLength(fReply,3);
 fReply[1]:=Chr(WhereY);
 fReply[2]:=Chr(WhereX);
 fReply[3]:=Chr(AID_CODE);
 for i:=0 to fFields.Count-1 do begin
  with TField(fFields[i]) do begin
   if fUpdated then begin
    fReply:=fReply+#$11; //  SBA
    fReply:=fReply+Chr(Top+1);
    fReply:=fReply+Chr(Left+1);
    fReply:=fReply+fText; // Graham change something here for Negative Signed Numeric Field...
   end;
  end;
 end;
 fOpCode:=3;
 fFlags:=0;
 SendReply;
end;

procedure TFrame5250.ATTNRequest;
begin
 fReply:='';
 fFlags:=$40;
 fOpCode:=0;
 SendReply;
end;

procedure TFrame5250.SystemRequest;
begin
 fReply:='';
 try
  if DialogBoxParam(hInstance,'SYSREQ',Handle,@SysReq,longint(Self))=2 then exit;
  fFlags:=$04;
  fOpCode:=0;
  AnsiToEbcdicA(fReply);
  SendReply;
 finally
  SetFocus;
 end;
end;

end.
