unit TN5250;

{
 Telnet 5250 emulation (C) 1999-2004, by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

{

RFC 1205 :

   An example of a typical negotiation process to establish 5250 mode of
   operation is shown below.  In this example, the server initiates the
   negotiation by sending the DO TERMINAL-TYPE request.

    Server: IAC DO TERMINAL-TYPE
    Client: IAC WILL TERMINAL-TYPE
    Server: IAC SB TERMINAL-TYPE SEND IAC SE
    Client: IAC SB TERMINAL-TYPE IS IBM-5251-11 IAC SE
    (The client has specified its terminal-type is an IBM-5251-11)

     IBM-5555-C01   24 x 80 Double-Byte Character Set color display
     IBM-5555-B01   24 x 80 Double-Byte Character Set (DBCS)
     IBM-3477-FC    27 x 132 color display
     IBM-3477-FG    27 x 132 monochrome display
     IBM-3180-2     27 x 132 monochrome display
     IBM-3179-2     24 x 80 color display
     IBM-3196-A1    24 x 80 monochrome display
     IBM-5292-2     24 x 80 color display
     IBM-5291-1     24 x 80 monochrome display
     IBM-5251-11    24 x 80 monochrome display

    Server: IAC DO END-OF-RECORD
    Client: IAC WILL END-OF-RECORD
    Server: IAC WILL END-OF-RECORD
    Client: IAC DO END-OF-RECORD
    (The server and client have both agreed to transmit EORs)

    Server: IAC DO TRANSMIT-BINARY
    Client: IAC WILL TRANSMIT-BINARY
    Server: IAC WILL TRANSMIT-BINARY
    Client: IAC DO TRANSMIT-BINARY
    (The server and client have both agreed to binary transmission)

}


type
  THead5250=packed record
   Size:word; // head + data
   Kind:word; // $A012 - WARNING ! can change in future version
   Zero:word; // Reserved
   Head:byte; // 4     - WARNING ! can change in future version
   Flag:word; // set of (ERR,ATN,f4,f8,SysRQ,TestRQ,HLP,?);
   Code:byte; // opcode
  end;

const
{ RFC 854
   In summary, WILL XXX is sent, by either party, to indicate that
   party's desire (offer) to begin performing option XXX, DO XXX and
   DON'T XXX being its positive and negative acknowledgments; similarly,
   DO XXX is sent to indicate a desire (request) that the other party
   (i.e., the recipient of the DO) begin performing option XXX, WILL XXX
   and WON'T XXX being the positive and negative acknowledgments.  Since
   the NVT is what is left when no options are enabled, the DON'T and
   WON'T responses are guaranteed to leave the connection in a state
   which both ends can handle.  Thus, all hosts may implement their
   TELNET processes to be totally unaware of options that are not
   supported, simply returning a rejection to (i.e., refusing) any
   option request that cannot be understood.
}
 TERM  = //'IBM-3179-2';
         'IBM-3477-FC';
       //'IBM-5251-11';

 tnEND      =#239; // End of 5250 DataStream
 tnSE       =#240;
 tnNOP      =#241;
 tnDATA     =#242;
 tnBRK      =#243;
 tnINT      =#244;
 tnABORT    =#245;
 tnAYT      =#246; // Are you there
 tnERASECHAR=#247;
 tnERASELINE=#248;
 tnGOAHEAD  =#249;
 tnSB       =#250;
 tnWILL     =#251;
 tnWONT     =#252;
 tnDO       =#253;
 tnDONT     =#254;
 tnIAC      =#255;

 tnBIN        =#0;
 tnECHO       =#1;
 tnSUPPRESS_GA=#3;
 tnTIMING     =#6;

 tnTERMTYPE   =#24; // Terminal type
 tnEOR        =#25;
 tnNEWENV     =#39; // Session name

 ttIS      =#0;
 ttSEND    =#1;
 ttINFO    =#2;

 tnVAR     =#0;
 tnVAL     =#1;

{
 Sample VT100 conversation :
 ---------------------------
 IAC DO NEWENV               - IAC WONT NEWENV
 IAC DO TERMTYPE             - IAC WILL TERMTYPE
 IAC SB TERMTYPE SEND IAC SE - IAC SB TERMTYPE IS "VT100" IAC SE
 IAC WILL ECHO               - IAC DO ECHO
 IAC WILL SUPPRESS_GA        - IAC DO SUPPRESS_GA

 Sample 5250 conversation :
 ---------------------------
 IAC DO NEWENV               - IAC WILL NEWENV
 IAC DO TERMTYPE             - IAC WILL TERMTYPE
 IAC SB NEWENV SEND IAC SE   - IAC SB NEWENV IS VAR "DEVNAME" VAL "TN5250" IAC SE
 IAC SB TERMTYPE SEND IAC SE - IAC SB TERMTYPE IS "IBM-5251-11" IAC SE
 IAC DO EOR                  - IAC WILL EOR
 IAC WILL EOR                - IAC DO EOR
 IAC DO BIN                  - IAC WILL BIN
 IAC WILL BIN                - IAC DO BIN
 <5250 Header & Packet>

}

implementation

end.
